@@ -8,7 +8,9 @@
  *	This program also contains some swiss-army-knife mailinglist	*
  *	support features.						*
  *									*
- *	Most notably:	flist	A program that should be setuid root.	*
+ *	Most notably:	flist	A program that should be setuid root	*
+ *	or setuid to the list user and setgid to the list user's	*
+ *	primary group.							*
  *									*
  *	Seems to be relatively bug free.				*
  *									*
@@ -17,9 +19,9 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: multigram.c,v 1.90 1999/03/30 18:17:40 guenther Exp $";
+ "$Id: multigram.c,v 1.94 2000/06/12 04:52:36 guenther Exp $";
 #endif
-static /*const*/char rcsdate[]="$Date: 1999/03/30 18:17:40 $";
+static /*const*/char rcsdate[]="$Date: 2000/06/12 04:52:36 $";
 #include "includes.h"
 #include "sublib.h"
 #include "hsort.h"
@@ -296,7 +298,7 @@ dble_gram:;
   return meter;
 }
 
-main(argc,argv)int argc;char*argv[];
+int main(argc,argv)int argc;char*argv[];
 { struct string fuzzstr,hardstr,excstr,exc2str;FILE*hardfile,**hfiles;
   const char*addit=0,*ldomain=0;char**nargv;size_t lldomain;
   struct match{char*fuzz,*hard;int metric;long lentry;off_t offs1,offs2;
@@ -340,8 +342,9 @@ Usage: flist listname[-request]\n\
 	     /*
 	      * continue as the compile-time-determined list maintainer
 	      */
-	      setgid(pass->pw_gid);initgroups(listid,pass->pw_gid);
-	      setuid(pass->pw_uid);
+	      if(setgid(pass->pw_gid)||initgroups(listid,pass->pw_gid)||
+	       setuid(pass->pw_uid))		 /* this can fail on certain */
+		 return EX_OSERR;	 /* broken systems e.g. Linux 2.2.14 */
 	    }
 	   else if(!(pass=getpwuid(euid)))
 	    { nlog("Euid");fprintf(stderr," %d",(int)euid);
@@ -409,7 +412,9 @@ nochdir: { nlog("Couldn't chdir to");logqnl(chp);
     /*
      *	revoke any suid permissions now, since we're not flist
      */
-     setgid(getgid());setuid(getuid());
+     if((setgid(getgid())&&getgid()!=getegid())||
+      (setuid(getuid())&&getuid()!=geteuid()))
+	return EX_OSERR;		       /* this _really_ can't happen */
      if(ISPROGRAM(chp,idhash))				  /* idhash program? */
       { unsigned long hash=0;int i;
 	progname=idhash;
@@ -779,6 +784,8 @@ usg:
      ;{ char*chp;
 	static const char tpunctuation[]="@\\/!#$%^&*-_=+|~`';:,.?{}";
 #define punctuation	(tpunctuation+3)
+	static const char colonpunct[]="/:@!";
+#define routepunct	(colonpunct+1)
 	chp=fuzzstr.text;
 	if(!dodomain)			    /* still have to do with domain? */
 	 { char*echp=strchr(chp,'\0')-1;
@@ -789,7 +796,7 @@ usg:
 	   while(*chp&&strchr(punctuation,*chp))
 	      chp++;				/* strip leading punctuation */
 	   ;{ const char*colon;				/* no decnet address */
-	      if(*(colon=chp+strcspn(chp,":@!/"))==':'&&colon[1]!=':')
+	      if(*(colon=chp+strcspn(chp,colonpunct))==':'&&colon[1]!=':')
 		 chp=(char*)colon+1;	       /* strip leading ...: garbage */
 	    }
 	   while(echp>=chp&&strchr(tpunctuation,*echp))
@@ -809,10 +816,15 @@ usg:
 	       strstr(chp,".."))&&
 	      !(*chp=='<'&&*echp=='>')&&	  /* RFC-822 machine literal */
 	      !(incomplete&&strchr(chp,'.')))		      /* domain name */
-	    { if(lastfrom<0)
+reject:	    { if(lastfrom<0)
 		 lastfrom=!strcmp(SHFROM,chp);
 	      continue;			 /* apparently not an e-mail address */
 	    }
+	   ;{ const char*colon;
+	      if((*chp=='@'||*chp=='<'&&chp[1]=='@')&&	 /* leading at's are */
+		 (!(colon=strchr(chp,':'))||strchr(routepunct,colon[1])))
+		 goto reject;		  /* only allowed on route addresses */
+	    }
 	   lastfrom=0;tmemmove(fuzzstr.text,chp,echp-chp+2);
 	   checkparens('(',')',fuzzstr.text,echp);
 	   checkparens('[',']',fuzzstr.text,strchr(fuzzstr.text,'\0'));
